# Copyright 2022 by Contributors

import numpy as np
import pandas as pd
import plotly.io as pio
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from PDSUtilities.pandas import get_numerical_columns
from PDSUtilities.plotly import apply_default
from PDSUtilities.plotly import get_font
from PDSUtilities.plotly import get_marker
from PDSUtilities.plotly import get_colors
from PDSUtilities.plotly import update_layout
from PDSUtilities.plotly import update_title
from PDSUtilities.plotly import update_width_and_height

def get_labels(columns, labels):
    if isinstance(labels, list):
        message = "Length of labels list must match length of columns list..."
        assert len(columns) == len(labels), message
        labels = { f"{columns[c]}": labels[c] for c in range(len(columns))}
    return labels

def get_correlation_label(correlations, columns, labels, row, col, precision, align = "middle"):
    BR = "<br />"
    col_label = labels.get(columns[col], columns[col])
    row_label = labels.get(columns[row], columns[row])
    cor_label = f"<b>{np.round(correlations.iloc[row, col], precision)}</b>"
    if align == "top":
        return "<span>" + cor_label + BR + col_label + BR + row_label + "</span>"
    if align == "bottom":
        return "<span>" + col_label + BR + row_label + BR + cor_label + "</span>"
    return "<span>" + col_label + BR + cor_label + BR + row_label + "</span>"

def get_center(values):
    return (0.5*(min(values) + max(values)))

def get_scatter_text(x, y, text, font = {}):
    # TODO: Need to test for more than just list...
    x = x if isinstance(x, list) else [x]
    y = y if isinstance(y, list) else [y]
    return go.Scatter(x = x, y = y, text = text, textfont = font,
        mode = 'text', showlegend = False, hoverinfo = "skip",
    )

def plot_correlation_scatter(df, target = None, columns = None, labels = {},
    width = None, height = None, title = None, precision = 4,
    template = None, colors = 0, marker = {},
    font = {}, axis_font = {}, tick_font = {}, legend_font = {}, hover_font = {},
    label_font = {}, title_font = {}):
    #
    font = apply_default(get_font(), font)
    axis_font = apply_default(font, axis_font)
    tick_font = apply_default(font, tick_font)
    label_font = apply_default(font, label_font)
    legend_font = apply_default(font, legend_font)
    #
    marker = apply_default(get_marker(), marker)
    #
    colors = get_colors(colors)
    columns = get_numerical_columns(df, columns, target)
    labels = get_labels(columns, labels)
    rows = [columns[c] for c in range(len(columns))]
    cols = [columns[c] for c in range(len(columns))]
    correlations = df[columns].corr()
    #
    values = [] if target is None else [value for value in df[target].unique()]
    values = [] if target is None else df[target].unique()
    #
    fig = make_subplots(rows = len(rows), cols = len(cols),
        horizontal_spacing = 0.1/len(cols),
        vertical_spacing = 0.1/len(rows),
        shared_xaxes = True,
        shared_yaxes = True,
        # print_grid = True,
    )
    for r in range(len(rows)):
        for c in range(r):
            text = get_correlation_label(
                correlations, columns, labels, r, c, precision, "middle"
            )
            for value in values:
                selection = df[target] == value
                fig.append_trace(
                    go.Scatter(
                        x = df[selection][cols[c]],
                        y = df[selection][rows[r]],
                        mode = 'markers',
                        marker = get_marker(marker, color = colors[value]),
                        name = labels.get(target, target) + " = " + str(value),
                        legendgroup = target + " = " + str(value),
                        hoverlabel = dict(font = hover_font),
                        showlegend = r == 1 and c == 0,
                    ),
                    r + 1, c + 1
                )
            if target is None:
                fig.append_trace(
                    go.Scatter(
                        x = df[cols[c]],
                        y = df[rows[r]],
                        mode = 'markers',
                        marker = get_marker(marker, color = colors[0]),
                        name = rows[r] + "/" + cols[c],
                        hoverlabel = dict(font = hover_font),
                        showlegend = False,
                    ),
                    r + 1, c + 1
                )
            # Used to center correlation text in
            # the plot as plotly annotations...
            fig.append_trace(
                get_scatter_text(
                    get_center(df[cols[c]]), get_center(df[rows[r]]), text, label_font
                ),
                c + 1, r + 1
            )
    # Point axes in upper plots to the axes
    # in the corresponding lower plots...
    for r in range(len(rows)):
        for c in range(r):
            # (r, c) corresponds to who we are pointing at...
            x, y = (len(rows) - 1)*len(cols) + c, r*len(cols)
            # So (c, r) is who we are...
            fig.update_xaxes(matches = f"x{x+1}", row = c + 1, col = r + 1)
            fig.update_yaxes(matches = f"y{y+1}", row = c + 1, col = r + 1)
    for r in range(len(rows)):
        fig.update_yaxes(
            title_text = labels.get(rows[r], rows[r]), row = r + 1, col = 1
        )
    for c in range(len(cols)):
        fig.update_xaxes(
            title_text = labels.get(cols[c], cols[c]), row = len(rows), col = c + 1
        )
    for r in range(len(rows)):
        for c in range(r, len(cols)):
            fig.update_xaxes(showgrid = False, row = r + 1, col = c + 1)
            fig.update_yaxes(showgrid = False, row = r + 1, col = c + 1)
        for c in range(1, len(cols)):
            fig.update_yaxes(ticks = "", row = r + 1, col = c + 1)
    for r in range(len(rows) - 1):
        for c in range(len(cols)):
            fig.update_xaxes(ticks = "", row = r + 1, col = c + 1)
    fig.update_xaxes(title_font= axis_font, tickfont = tick_font, linecolor = "black")
    fig.update_yaxes(title_font= axis_font, tickfont = tick_font, linecolor = "black")
    fig.update_xaxes(linewidth = 0.5, mirror = True, zeroline = False)
    fig.update_yaxes(linewidth = 0.5, mirror = True, zeroline = False)
    #
    fig.update_layout(legend_font = legend_font)
    if target is not None:
        fig.update_layout(legend_itemsizing = 'constant')
        fig.update_layout(legend = dict(
            orientation = 'h', yanchor = 'top', xanchor = 'center', y = 1.07, x = 0.5
        ))
    #
    fig = update_width_and_height(fig, width, height)
    fig = update_title(fig, title, title_font, font)
    fig = update_layout(fig, font = font, template = template)
    return fig


class ColorblindSafeColormaps:
    """
     A collection of colorblind-safe colormaps.
    """
    def __init__(self):
        """
        __init__ create and initialize a ColorblindSafeColormaps object.
        """
        self.names = ["Vibrant", "Bright", "Muted", "Medium-Contrast", "Grayscale"]
        self.colors = [
            ["#0077BB", "#CC3311", "#33BBEE", "#EE7733", "#009988", "#EE3377", "#BBBBBB"],
            ["#4477AA", "#AA3377", "#66CCEE", "#EE6677", "#CCBB44", "#228833", "#BBBBBB"],
            ["#332288", "#882255", "#88CCEE", "#CC6677", "#44AA99", "#AA4499", "#117733",
            "#DDCC77", "#999933", "#DDDDDD"],
            ["#004488", "#994455", "#997700", "#6699CC", "#EE99AA", "#EECC66"],
            ["#444444", "#999999", "#666666", "#DDDDDD"],
        ]
    def get_names(self):
        """
        get_names returns a list of colormap names

        Returns:
            list of str: returns a list of colormap names
        """
        return self.names
    def get_count(self):
        """
        get_count returns the number of colormaps available

        Returns:
            int: the number of colormaps available
        """
        return len(self.names)
    def get_colors_by_index(self, index):
        """
        get_colors_by_index returns the colormap associated with index

        Args:
            index (int): the index of the colormap

        Returns:
            list of str: the list of colors in the colormap
        """
        return self.colors[index]
    def get_colors_by_name(self, name):
        """
        get_colors_by_name returns the colormap associated with name

        Args:
            name (str): the name of the colormap

        Returns:
            list of str: the list of colors in the colormap
        """
        return self.colors[self.names.index(name)]
    def get_colors(self, colors):
        """
        get_colors returns the colormap associated with name or index

        Args:
            name (str or int): the name or index of the colormap

        Returns:
            list of str: the list of colors in the colormap
        """
        if isinstance(colors, int):
            return self.get_colors_by_index(colors)
        return self.get_colors_by_name(colors)

if __name__ == "__main__":
   colormaps = ColorblindSafeColormaps()
   print("Number of colormaps: ", colormaps.get_count())
   print("Names of colormaps: ", colormaps.get_names())
   print(colormaps.get_colors_by_index(0))
   print(colormaps.get_colors_by_name("Vibrant"))
   print(colormaps.get_colors_by_index(-1))
   print(colormaps.get_colors_by_name("Grayscale"))
   print(colormaps.get_colors(0))
   print(colormaps.get_colors("Vibrant"))




from PDSUtilities.pandas.plot_histograms import plot_histograms
from PDSUtilities.pandas.plot_parallel_categories import plot_parallel_categories
from PDSUtilities.pandas.plot_parallel_coordinates import plot_parallel_coordinates
from PDSUtilities.pandas.plot_parallel_coordinates import get_numerical_columns
from PDSUtilities.pandas.plot_parallel_categories import get_categorical_columns
from PDSUtilities.pandas.plot_correlation_scatter import plot_correlation_scatter
from PDSUtilities.pandas.plot_correlation_matrix import plot_correlation_matrix
from PDSUtilities.pandas.plot_correlation_triangle import plot_correlation_triangle



from ipywidgets import Button, Dropdown

def prev_dropdown_next(description, options, on_change):
    def get_prev_next_functions(dropdown):
        def on_prev(b):
            dropdown.value = (dropdown.value - 1) % len(dropdown.options)
        def on_next(b):
            dropdown.value = (dropdown.value + 1) % len(dropdown.options)
        return on_prev, on_next
    dropdown = Dropdown(
        description = description,
        value = 0,
        options = options
    )
    dropdown.observe(on_change, names = "value")
    prev_button = Button(description = "< Prev")
    next_button = Button(description = "Next >")
    on_prev, on_next = get_prev_next_functions(dropdown)
    prev_button.on_click(on_prev)
    next_button.on_click(on_next)
    return prev_button, dropdown, next_button



from PDSUtilities.plotly.ColorblindSafeColormaps import ColorblindSafeColormaps
from PDSUtilities.plotly.utilities import apply_default
from PDSUtilities.plotly.utilities import get_font
from PDSUtilities.plotly.utilities import get_shape
from PDSUtilities.plotly.utilities import get_line
from PDSUtilities.plotly.utilities import get_arrow
from PDSUtilities.plotly.utilities import get_label
from PDSUtilities.plotly.utilities import get_marker
from PDSUtilities.plotly.utilities import update_layout
from PDSUtilities.plotly.utilities import hex_to_rgb
from PDSUtilities.plotly.utilities import rgb_to_hex
from PDSUtilities.plotly.utilities import get_colors
from PDSUtilities.plotly.utilities import update_width_and_height
from PDSUtilities.plotly.utilities import update_title
from PDSUtilities.plotly.utilities import remove_ticks_and_tick_labels
from PDSUtilities.plotly.create_image_subplots import create_image_subplots
import PDSUtilities.plotly.templates as templates



from PDSUtilities.plotly import ColorblindSafeColormaps

DEFAULT_FONT = {
	'family': "Verdana, Helvetica, Verdana, Calibri, Garamond, Cambria, Arial",
	'size': 16,
	'color': "#000000"
}
DEFAULT_SHAPE = {
	'type': "rect",
	'fillcolor': "#FFFFFF",
	'opacity': 1.0,
}
DEFAULT_LINE = {
	'color': "#000000",
	'width': 1,
	# ['solid', 'dot', 'dash', 'longdash', 'dashdot', 'longdashdot']
	'dash': "solid",
}
DEFAULT_ARROW = {
    # Integer between or equal to 0 and 8
	'arrowhead': 0,
	# Relative to arrowwidth
	'arrowsize': 1,
	'arrowwidth': 1,
}
DEFAULT_LABEL = {
	'align': "center",
	'bgcolor': "#FFFFFF",
	'bordercolor': "rgba(0,0,0,0)",
	'borderpad': 1,
	'borderwidth': 1,
	'opacity': 1.0,
	'textangle': 0,
	'valign': "middle",
	'visible': True,
}

DEFAULT_MARKER = dict(
    size = 2,
    color = "#CCCCCC",
)

def apply_default(old_thing, new_thing):
    return { **old_thing, **new_thing } if new_thing is not None else old_thing

def get_font(font = None, family = None, size = None, color = None):
    font = apply_default(DEFAULT_FONT, font)
    font['family'] = family if family is not None else font['family']
    font['size'] = size if size is not None else font['size']
    font['color'] = color if color is not None else font['color']
    return font

def get_shape(shape = None, type = None, fillcolor = None, opacity = None):
    shape = apply_default(DEFAULT_SHAPE, shape)
    shape['type'] = type if type is not None else shape['type']
    shape['fillcolor'] = fillcolor if fillcolor is not None else shape['fillcolor']
    shape['opacity'] = opacity if opacity is not None else shape['opacity']
    return shape

def get_line(line = None, width = None, color = None, dash = None):
    line = apply_default(DEFAULT_LINE, line)
    line['width'] = width if width is not None else line['width']
    line['color'] = color if color is not None else line['color']
    line['dash'] = dash if dash is not None else line['dash']
    return line

def get_arrow(arrow = None, arrowhead = None, arrowsize = None, arrowwidth = None):
    arrow = apply_default(DEFAULT_ARROW, arrow)
    arrow['arrowhead'] = arrowhead if arrowhead is not None else arrow['arrowhead']
    arrow['arrowsize'] = arrowsize if arrowsize is not None else arrow['arrowsize']
    arrow['arrowwidth'] = arrowhead if arrowwidth is not None else arrow['arrowwidth']
    return arrow

# Tweak this API as needed...
def get_label(label = None):
    label = apply_default(DEFAULT_LABEL, label)
    # label['bgcolor'] = bgcolor if bgcolor is not None else label['bgcolor']
    return label

# Tweak this API as needed...
def get_marker(marker = None, size = None, color = None):
    marker = apply_default(DEFAULT_MARKER, marker)
    marker['size'] = size if size is not None else marker['size']
    marker['color'] = color if color is not None else marker['color']
    return marker

def update_layout(fig, font = {}, template = None):
    fig.update_layout(font = font)
    if template is not None:
        fig.update(template = template)
    return fig

def hex_to_rgb(value):
    value = value.lstrip('#')
    lv = len(value)
    return tuple(int(value[i:i + lv // 3], 16) for i in range(0, lv, lv // 3))

def rgb_to_hex(rgb):
    return "#%02x%02x%02x" % rgb

def get_colors(colors, default = 0):
    colors = default if colors is None else colors
    if isinstance(colors, int):
        colors = ColorblindSafeColormaps().get_colors(colors)
    return colors

def update_width_and_height(fig, width = None, height = None):
    if width is not None:
        fig.update_layout(width = width)
    if height is not None:
        fig.update_layout(height = height)
    return fig

def update_title(fig, title, title_font = {}, font = {}):
    title_font = apply_default(
        apply_default(font, { 'size': font.get('size', 16) + 4 }),
        title_font
    )
    if title is not None and isinstance(title, str):
        title = { 'text': title, 'x': 0.5, 'xanchor': "center" }
    if title is not None:
        fig.update_layout(title = title, title_font = title_font)
    return fig

def remove_ticks_and_tick_labels(fig, rows = None, cols = None):
    if rows is not None and cols is not None:
        for row in range(rows):
            for col in range(cols):
                fig.update_xaxes(row = row + 1, col = col + 1,
                    ticks = "", tickfont_size = 1, tickfont_color = "rgba(0,0,0,0)"
                )
                fig.update_yaxes(row = row + 1, col = col + 1,
                    ticks = "", tickfont_size = 1, tickfont_color = "rgba(0,0,0,0)"
                )
    else:
        fig.update_xaxes(
            ticks = "", tickfont_size = 1, tickfont_color = "rgba(0,0,0,0)"
        )
        fig.update_yaxes(
            ticks = "", tickfont_size = 1, tickfont_color = "rgba(0,0,0,0)"
        )
    return fig



import numpy as np
import plotly.graph_objects as go
import plotly.io as pio
from plotly.subplots import make_subplots

def create_image_subplots(images, rows, cols, width = None, height = None, template = None):
    fig = make_subplots(rows = rows, cols = cols,
        shared_xaxes = False,
        shared_yaxes = False,
        vertical_spacing = 0.01,
        horizontal_spacing = 0.01,
    )
    for row in range(rows):
        for col in range(cols):
            index = row*cols + col
            fig.add_trace(go.Image(z = images[index]), row = 1 + row, col = 1 + col)
    _, max_y, max_x, _ = np.shape(images)
    height = height if height is not None else 100 + 200*rows
    width = width if width is not None else 200*cols
    for row in range(rows):
        for col in range(cols):
            fig.update_xaxes(row = row + 1, col = col + 1,
                range = [0, max_x - 1], showgrid = False, ticks = "",
                tickfont_size = 1, tickfont_color = "#FFFFFF",
                linecolor = "black", linewidth = 0.5, zeroline = False
            )
            fig.update_yaxes(row = row + 1, col = col + 1,
                range = [0, max_y - 1], showgrid = False, ticks = "",
                tickfont_size = 1, tickfont_color = "#FFFFFF",
                linecolor = "black", linewidth = 0.5, zeroline = False
            )
    fig.update_layout(height = height, width = width)
    if template is not None:
        fig.update_layout(template = template)
    return fig




import plotly.graph_objects as go
import plotly.io as pio

# Template settings for plotly...
# https://plotly.com/python/templates/

layout_axis = dict(
    mirror=True,
    ticks="outside",
    showline=True,
    title_standoff=5,
    showgrid=True,
    zeroline=False,
)

pio.templates["DrJohnWagner"] = go.layout.Template(
    layout_xaxis=layout_axis,
    layout_yaxis=layout_axis,
    layout_title_font_size=18,
    layout_font_family="Verdana, Helvetica, Verdana, Calibri, Garamond, Cambria, Arial",
    layout_font_size=16,
)

pio.templates.default = "DrJohnWagner"

pio.templates["draft"] = go.layout.Template(
    layout_annotations=[
        dict(
            name="draft watermark",
            text="DRAFT",
            textangle=-30,
            opacity=0.1,
            font=dict(color="black", size=100),
            xref="paper",
            yref="paper",
            x=0.5,
            y=0.5,
            showarrow=False,
        )
    ]
)



import os
import numpy as np

import tensorflow as tf
from tensorflow.keras.models import load_model

import plotly.graph_objects as go
import plotly.io as pio

from ipywidgets import widgets
from ipywidgets import interactive, HBox, VBox, Button

from PDSUtilities.ipywidgets import prev_dropdown_next
from PDSUtilities.plotly import create_image_subplots

from utilities import load_image

def scale_image(image):
# 	image = image.numpy()
# 	image = image[..., 0]
    numerator = np.max(image) - np.min(image)
    numerator = 1.0 if numerator < 0.001 else numerator
    image = 255*(image - np.min(image))/numerator
    image = image.astype(int)
    return np.stack([image, image, image], axis = -1)

def cast_image(image):
    image = tf.expand_dims(image, axis=-1)
    image = tf.cast(image, tf.float32)
    image = 2.0*image/255.0 - 1.0
    return image

def view_model(model, filenames, rows, cols, template):
    if isinstance(model, str):
        model = load_model(model)
    if template is None:
        template = pio.templates.default
    if template is None:
        template = go.layout.Template()

    def get_options(layer):
        layer = model.layers[layer]
        # assert len(layer.output.shape) == 4
        filters, rc = layer.output.shape[-1], rows*cols
        return [
            (f"Filters {f} to {min(f + rc, filters) - 1}", f//rc)
            for f in range(0, filters, rc)
        ]
    # For now just get first four...
    def get_filters(image, layer, filter):
        #image = os.path.join("./training/x", image)
        image = cast_image(load_image(image, True))
        layer = model.layers[layer]
        filters = tf.keras.Model(
            inputs = model.inputs,
            outputs = layer.output
        ).predict(np.asarray([image]))
        lo = filter*rows*cols
        hi = min(lo + rows*cols, layer.output.shape[-1])
        return [
            filters[0][..., f] for f in range(lo, hi)
        ] + [
            np.zeros_like(filters[0][..., 0]) for f in range(hi, lo + rows*cols)
        ]

    def on_figure(image, layer, filter):
        filters = get_filters(image, layer, filter)
        _, height, width = np.shape(filters)
        with fig.batch_update():
            for f, filter in enumerate(filters):
                fig.data[f].z = scale_image(filter)
        for row in range(rows):
            for col in range(cols):
                fig.update_xaxes(
                    row = row + 1, col = col + 1,
                    range = [0, width - 1],
                )
                fig.update_yaxes(
                    row = row + 1, col = col + 1,
                    range = [0, height - 1],
                )

    def on_image(change):
        image, _ = images.options[change['new']]
        layer, filter = layers.value, filters.value
        on_figure(image, layer, filter)
    prev_image, images, next_image = prev_dropdown_next(
        "Images", [
            (filename, f) for f, filename in enumerate(filenames)
        ], on_image
    )
    #
    def on_layer(change):
        filters.value = 0
        filters.options = get_options(change['new'])
        image, _ = images.options[images.value]
        on_figure(image, change['new'], 0)
    prev_layer, layers, next_layer = prev_dropdown_next(
        "Layers",
        [(layer.name, l) for l, layer in enumerate(model.layers)],
        on_layer
    )
    #
    def on_filter(change):
        image, _ = images.options[images.value]
        on_figure(image, layers.value, change['new'])
    prev_filter, filters, next_filter = prev_dropdown_next(
        "Filters", get_options(layers.value), on_filter
    )
    #
    def create_figure(image, layer, filter, rows, cols, template):
        filters = get_filters(image, layer, filter)
        filters = [scale_image(filter) for filter in filters]
        return create_image_subplots(filters, rows, cols, template = template)
    fig = go.FigureWidget(
        create_figure(
            images.options[images.value][0],
            layers.value, filters.value, rows, cols, template
        ),
        layout=go.Layout(title = "Title"),
    )

    return widgets.VBox([
        widgets.HBox([images, prev_image, next_image]),
        widgets.HBox([layers, prev_layer, next_layer]),
        widgets.HBox([filters, prev_filter, next_filter]),
        fig
    ])


# Copyright 2022 by Contributors

import numpy as np
import pandas as pd
import plotly.io as pio
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from PDSUtilities.plotly import apply_default
from PDSUtilities.plotly import get_font
from PDSUtilities.plotly import ColorblindSafeColormaps

def get_categories_and_counts(df, column, target, value):
    categories = df[column].unique()
    df = df[df[target] == value]
    counts = [df[df[column] == category][column].count() for category in categories]
    return categories, counts

# Plotly is too smart and converts strings to numbers when
# possible but we're smarter: wrap numbers in <span></span>!
def to_string(value):
    if isinstance(value, str):
        return value
    return f"<span>{value}</span>"

def get_categories_and_counts(df, column, target = None, value = None):
    categories = df[column].unique()
    if target is not None:
        df = df[df[target] == value]
    counts = [df[df[column] == category][column].count() for category in categories]
    categories = [to_string(category) for category in categories]
    return categories, counts

def get_width(index):
    WIDTHS = [0.0, 0.1, 0.23, 0.34, 0.46, 0.65, 0.8]
    width = WIDTHS[index] if index < len(WIDTHS) else WIDTHS[-1]
    return width

def get_bins(bins, column):
    if isinstance(bins, int):
        return bins
    return bins.get(column, 0)

def get_histogram(x, color, show_legend, name, cumulative, bins):
    return go.Histogram(
        x = x, marker_color = color, showlegend = show_legend,
        name = name, cumulative_enabled = cumulative, nbinsx = bins,
        legendgroup = name,
    )

def get_bar(categories, counts, color, show_legend, name):
    return go.Bar(
        x = categories, y = counts,
        marker_color = color,
        showlegend = show_legend,
        name = name,
        width = get_width(len(categories)),
        legendgroup = name,
    )

def get_rcwh(rows, cols, width, height, columns, values):
    columns = len(columns)
    w, h = 250, 200
    if rows is None and cols is None:
        cols = max(2, min(5, int(np.ceil(np.sqrt(columns)))))
        rows = int(np.ceil(columns/cols))
    elif cols is None:
        cols = int(np.ceil(columns/rows))
    elif rows is None:
        rows = int(np.ceil(columns/cols))
    if width is None:
        if cols > 2 or cols == 2 and len(values) < 4:
            width = w*cols
        else:
            width = w*(cols + 1)
    if height is None:
        if cols > 2 or cols == 2 and len(values) < 4:
            height = 100 + h*rows
        else:
            height = h*rows
    return rows, cols, width, height

# produces histograms for all categorical and numerical columns in a pandas dataframe
def plot_histograms(df, target = None, columns = None, rows = None, cols = None, width = None, height = None,
    title = None, cumulative = None, barmode = "stack", opacity = 0.65, bins = 0,
    hovermode = "x unified", template = None, colors = 0, font = {}, title_font = {}, legend_font = {}):
    """
    plot_histograms produces histograms for all categorical and numerical columns in a pandas dataframe.

    Args:
        df (dataframe): the pandas dataframe to plot historgrams of.
        target (str, optional): the name of the target column whose values are used for grouping.
            Defaults to None.
        columns (list[str], optional): a list of column names to plot. Defaults to all columns.
        rows (int, optional): the number of rows in the grid the histograms are arranged in.
            Defaults to None.
        cols (int, optional): the number of columns in the grid the histograms are arranged in.
            Defaults to None.
        width (int, optional): the plot's width in pixels. Defaults to None.
        height (int, optional): the plot's height in pixels. Defaults to None.
        title (str or dict, optional): the title displayed at the top of the figure.
            Defaults to None.
        cumulative (bool, optional): produces cumulative histograms for each target variable, or the
            entire dataset if target = None. Defaults to None.
        barmode (str, optional): specifies whether the bars should be stacked, grouped or overlaid.
            Defaults to "stack".
        opacity (float, optional): set's the opacity of the bars when barmode = "overlay".
            Defaults to 0.65.
        bins (int or dict, optional): the number of bins for numerical colum histograms.
            Defaults to 0.
        hovermode (str, optional): specifies the style of hover popups.
            Defaults to "x unified".
        template (str, optional): the name of a plotly template to apply to the figure.
            Defaults to None.
        colors (int, str or list, optional): specifies the colors to be used for the bars.
            Defaults to 0.
        font (dict, optional): the font to be used in the histograms. Defaults to the default font.
        title_font (dict, optional): the font to be used for the figure title. Defaults to font.
        legend_font (dict, optional): the font to be used for the legend. Defaults to font.

    Returns:
        plotly.graph_objects.Figure: the plotly Figure encapsulating the histogram plots.
    """
    default_font = get_font()
    font = apply_default(default_font, font)
    legend_font = apply_default(font, legend_font)
    title_font = apply_default(
        apply_default(font, { 'size': font.get('size', 16) + 4 }),
        title_font
    )
    colors = 0 if colors is None else colors
    if isinstance(colors, int):
        colors = ColorblindSafeColormaps().get_colors(colors)
    #
    values = [] if target is None else [value for value in df[target].unique()]
        #
    if columns is None:
        columns = [column for column in df.columns if column != target]
    if not isinstance(columns, list):
        columns = [column for column in columns]
    if target is not None and target in columns:
        columns.remove(target)
    #
    rows, cols, width, height = get_rcwh(rows, cols, width, height, columns, values)
    fig = make_subplots(rows = rows, cols = cols,
        horizontal_spacing = 0.25/cols,
        vertical_spacing = 0.37/rows,
        subplot_titles = columns,
    )
    for index, column in enumerate(columns):
        for value in values:
            name = f"{target} = {value}"
            max_bins = get_bins(bins, column)
            color = colors[values.index(value) % len(colors)]
            if df[column].dtypes == object or len(df[column].unique()) <= len(colors):
                categories, counts = get_categories_and_counts(df, column, target, value)
                trace = get_bar(categories, counts, color, index == 0, name)
                fig.append_trace(trace, 1 + index // cols, 1 + index % cols)
            else:
                x = df[df[target] == value][column]
                trace = get_histogram(x, color, index == 0, name, cumulative, max_bins)
                fig.append_trace(trace, 1 + index // cols, 1 + index % cols)
        if target is None:
            if df[column].dtypes == object or len(df[column].unique()) <= len(colors):
                categories, counts = get_categories_and_counts(df, column)
                trace = get_bar(categories, counts, colors[0], False, column)
                fig.append_trace(trace, 1 + index // cols, 1 + index % cols)
            else:
                trace = get_histogram(df[column], colors[0], False, column, cumulative, max_bins)
                fig.append_trace(trace, 1 + index // cols, 1 + index % cols)
    # barmode = ['stack', 'group', 'overlay', 'relative']
    # barmode = "stack"
    if barmode == "overlay":
        fig.update_traces(opacity = opacity)
    fig.update_annotations(font = font)
    fig.update_traces(marker_line_color = "#000000")
    fig.update_traces(marker_line_width = 0.5)
    if title is not None and isinstance(title, str):
        title = { 'text': title, 'x': 0.5, 'xanchor': "center" }
    if title is not None:
        fig.update_layout(title = title)
    if template is not None:
        fig.update_layout(template = template)
    if cols > 2 or cols == 2 and len(values) < 4:
        fig.update_layout(width = width, height = height, barmode = barmode,
            legend = dict(
                orientation = "h",
                yanchor = "bottom",
                y = 1.0 + 2.0*cols/100.0,
                xanchor = "center",
                x = 0.5
            ),
            margin = { 't': 160 },
        )
    fig.update_layout(hovermode = hovermode)
    fig.update_layout(width = width, height = height, barmode = barmode,
        font = font, title_font = title_font, legend_font = legend_font,
        # margin = { 't': 160 },
        # bargap = 0.2, # gap between bars of adjacent location coordinates
        # bargroupgap = 0*0.2, # gap between bars of the same location coordinates
    )
    # This is literally the dumbest thing I've seen in years...
    # This puts space between the ticks and tick labels. SMFH.
    fig.update_yaxes(ticksuffix = " ")
    return fig



# Copyright 2022 by Contributors

import numpy as np
import pandas as pd
import plotly.io as pio
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from PDSUtilities.plotly import apply_default
from PDSUtilities.plotly import get_font
from PDSUtilities.plotly import get_marker
from PDSUtilities.plotly import update_layout
from PDSUtilities.plotly import hex_to_rgb
from PDSUtilities.plotly import rgb_to_hex
from PDSUtilities.plotly import get_colors
from PDSUtilities.plotly import update_title
from PDSUtilities.plotly import update_width_and_height
from PDSUtilities.pandas import get_numerical_columns

def get_labels(columns, labels):
    if isinstance(labels, list):
        message = "Length of labels list must match length of columns list..."
        assert len(columns) == len(labels), message
        labels = { f"{columns[c]}": labels[c] for c in range(len(columns))}
    return labels

def get_color(value, colors):
    rlo, glo, blo = colors[1]
    rhi, ghi, bhi = colors[0]
    fraction = (value + 1.0)/2.001
    r = rlo + int(np.round(fraction*(rhi - rlo)))
    g = glo + int(np.round(fraction*(ghi - glo)))
    b = blo + int(np.round(fraction*(bhi - blo)))
    return rgb_to_hex((r, g, b))

def get_scatter_text(x, y, text, font = {}):
    # TODO: Need to test for more than just list...
    x = x if isinstance(x, list) else [x]
    y = y if isinstance(y, list) else [y]
    return go.Scatter(x = x, y = y, text = text, textfont = font,
        mode = 'text', showlegend = False, hoverinfo = "skip",
    )

# font
# axis_font
# hover_font
# label_font
# title_font
#
# tick_font
# legend_font
# subtitle_font
def plot_correlation_triangle(df, columns = None, labels = {},
    width = None, height = None, title = None, precision = 4,
    template = None, colors = 0, xangle = 45, yangle = 45,
    font = {}, axis_font = {}, hover_font = {}, label_font = {},
    title_font = {}):
    #
    font = apply_default(get_font(), font)
    axis_font = apply_default(font, axis_font)
    label_font = apply_default(font, label_font)
    label_font = apply_default(label_font, { 'color': "#FFFFFF" })
    #
    colors = get_colors(colors)
    colors = [hex_to_rgb(color) for color in colors]
    columns = get_numerical_columns(df, columns)
    labels = get_labels(columns, labels)
    rows = [columns[c] for c in range(1, len(columns)    )]
    cols = [columns[c] for c in range(0, len(columns) - 1)]
    correlations = df[columns].corr()
    #
    fig = make_subplots(rows = len(rows), cols = len(cols),
        horizontal_spacing = 0.1/len(cols),
        vertical_spacing = 0.1/len(rows),
        shared_xaxes = True,
        shared_yaxes = True,
        # print_grid = True,
    )
    for r in range(len(rows)):
        for c in range(r + 1):
            value = correlations[cols[c]][rows[r]]
            fig.append_trace(
                go.Scatter(
                    x = [0.0], y = [0.0],
                    mode = 'markers+text',
                    # Make a large square to fill the plot area
                    # since we can't set the background color...
                    marker = dict(
                        symbol = "square",
                        size = 1000,
                        color = get_color(value, colors),
                    ),
                    hoverlabel = dict(font = hover_font),
                    hovertemplate = f"{rows[r]}<br>{cols[c]}",
                    text = str(np.round(value, precision)),
                    name = str(np.round(value, precision)),
                    textfont = label_font,
                    showlegend = False,
                ),
                r + 1, c + 1
            )
    fig.update_xaxes(range = [-1.0, 1.0])
    fig.update_yaxes(range = [-1.0, 1.0])
    fig.update_xaxes(showgrid = False, ticks = "",  mirror = True)
    fig.update_yaxes(showgrid = False, ticks = "",  mirror = True)
    fig.update_xaxes(linecolor = "black", linewidth = 0.5, zeroline = False)
    fig.update_yaxes(linecolor = "black", linewidth = 0.5, zeroline = False)
    # We use a single tick label as the axis label...
    for c in range(len(cols)):
        fig.update_xaxes(
            tickmode = "array",
            tickvals = [0.0],
            tickfont = axis_font,
            ticktext = [labels.get(cols[c], cols[c])],
            tickangle = -xangle,
            row = len(rows), col = c + 1,
        )
    for r in range(len(rows)):
        fig.update_yaxes(
            tickmode = "array",
            tickvals = [0.2 if yangle == 45 else 0.0],
            tickfont = axis_font,
            ticktext = [
                labels.get(rows[r], rows[r]) + "  " if yangle == 45 else " "
            ],
            tickangle = -yangle,
            row = r + 1, col = 1,
        )
    #
    fig = update_width_and_height(fig, width, height)
    fig = update_title(fig, title, title_font, font)
    fig = update_layout(fig, font = font, template = template)
    return fig


from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("PDSUtilities/version.py") as f:
    exec(f.read())

setup(
    name="PDSUtilities",
    version=__version__,
    description="Utilities for data science in python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/DrJohnWagner/PDSUtilities",
    author="Dr. John Wagner",
    author_email="Dr.John.Wagner@gmail.com",
    license="Apache-2.0",
    packages=[
        "PDSUtilities",
        "PDSUtilities.ipywidgets",
        "PDSUtilities.pandas",
        "PDSUtilities.plotly",
        "PDSUtilities.tensorflow",
        "PDSUtilities.xgboost",
    ],
    install_requires=[
        "igraph",
        "ipywidgets",
        "nbformat",
        "numpy",
        "pandas",
        "plotly",
        "tensorflow",
        "xgboost",
    ],
    tests_require=[
        "pytest",
    ],
    zip_safe=False,
)

# package_dir={"": "src"},
# packages=setuptools.find_packages(where="src"),



# Copyright 2022 by Contributors

from xgboost import XGBModel
from xgboost import Booster
import plotly.graph_objects as go

# The `PDSUtilities.xgboost.plot_importance()` function is a direct
# copy/paste/edit modification of `xgboost.plot_importance()` with a few
# minor tweaks to the API and relatively light changes to the code. The
# xgboost team deserves the vast majority of credit for this code!
# The xgboost license can be found here:
# https://github.com/dmlc/xgboost/blob/master/LICENSE

# TODO: #1 change features to labels and update README.md.
def plot_importance(booster, labels = {}, width = 0.6, xrange = None, yrange = None,
    title = 'Feature Importance', xlabel = 'F Score', ylabel = 'Features', fmap = '',
    max_features = None, importance_type = 'weight', show_grid = True, show_values = True):
    """Plot importance based on fitted trees using plotly.
    Parameters
    ----------
    booster : Booster, XGBModel or dict
        Booster or XGBModel instance, or dict taken by Booster.get_fscore()
	labels: list or dict of labels for displaying feature names.
		* labels dict maps default feature names ("f0", "f1", etc) to labels
		* labels list is a list of labels corresponding by index to features
    show_grid : bool, Turn the axes grids on or off.  Default is True (On).
    importance_type : str, default "weight"
        How the importance is calculated: either "weight", "gain", or "cover"
        * "weight" is the number of times a feature appears in a tree
        * "gain" is the average gain of splits which use the feature
        * "cover" is the average coverage of splits which use the feature
          where coverage is defined as the number of samples affected by the split
    max_features : int, default None
        Maximum number of top features displayed on plot. If None, all features will be displayed.
    width : float, default 0.5
        Bar width
    xlimits : tuple, default None
        Tuple passed to axes.xlim()
    ylimits : tuple, default None
        Tuple passed to axes.ylim()
    title : str, default "Feature importance"
        Axes title. To disable, pass None.
    xlabel : str, default "F score"
        X axis title label. To disable, pass None.
    ylabel : str, default "Features"
        Y axis title label. To disable, pass None.
    fmap: str or os.PathLike (optional)
        The name of feature map file.
    show_values : bool, default True
        Show values on plot. To disable, pass False.
    Returns
    -------
    fig : plotly Figure object
    """
    try:
        import plotly.graph_objects as go
    except ImportError as e:
        raise ImportError('You must install plotly to plot importance') from e

    if isinstance(booster, XGBModel):
        importance = booster.get_booster().get_score(
            importance_type = importance_type, fmap = fmap)
    elif isinstance(booster, Booster):
        importance = booster.get_score(importance_type = importance_type, fmap = fmap)
    elif isinstance(booster, dict):
        importance = booster
    else:
        raise ValueError('tree must be Booster, XGBModel or dict instance')

    if not importance:
        raise ValueError(
            'Booster.get_score() results in empty.  ' +
            'This maybe caused by having all trees as decision dumps.')

    if isinstance(labels, list):
        labels = {
            f"f{feature}": labels[feature] for feature in range(len(labels))
        }

    tuples = [(k, importance[k]) for k in importance]
    if max_features is not None:
        # pylint: disable=invalid-unary-operand-type
        tuples = sorted(tuples, key=lambda x: x[1])[-max_features:]
    else:
        tuples = sorted(tuples, key=lambda x: x[1])
    labels, values = zip(*tuples)

    text = [xlabel + ": " + str(value) for value in values]
    fig = go.Figure(go.Bar(
            y = [labels.get(label, label.upper()) for label in labels],
            x = values,
            orientation = 'h',
            width = width,
            hovertext = text if show_values else [],
            text = text if show_values else [],
            textposition = 'auto',
        ))
    if xrange is not None:
        if not isinstance(xrange, tuple) or len(xrange) != 2:
            raise ValueError('xrange must be a tuple of 2 elements')
        fig.update_xaxes(range = xrange)
    if yrange is not None:
        if not isinstance(yrange, tuple) or len(yrange) != 2:
            raise ValueError('yrange must be a tuple of 2 elements')
        fig.update_yaxes(range = yrange)
    if title is not None:
        fig.update_layout(title = {"text": title, "x": 0.5, "xanchor":  "center"})
    if xlabel is not None:
        fig.update_xaxes(title_text = xlabel)
    if ylabel is not None:
        fig.update_yaxes(title_text = ylabel)
    fig.update_xaxes(showgrid = show_grid)
    fig.update_yaxes(showgrid = False)
    # This is literally the dumbest thing I've seen in years...
    # This puts space between the ticks and tick labels. SMFH.
    fig.update_yaxes(ticksuffix = "  ")
    return fig


# Copyright 2022 by Contributors

import numpy as np
import pandas as pd
import plotly.io as pio
import plotly.graph_objects as go
from PDSUtilities.plotly import get_colors
from PDSUtilities.plotly import apply_default
from PDSUtilities.plotly import get_font
from PDSUtilities.plotly import update_title
from PDSUtilities.plotly import update_width_and_height


def get_categorical_columns(df, columns=None, remove=None):
    if columns is None:
        columns = [column for column in df.columns if df[column].dtypes == "O"]
    if not isinstance(columns, list):
        columns = [column for column in columns]
    if remove is not None and remove in columns:
        columns.remove(remove)
    return columns


def get_line(df, target, colors):
    line = dict(
        color=colors[0],
        showscale=False,
    )
    if target is not None:
        values = df[target]
        if df[target].dtypes == "O":
            values = df[target].astype("category").cat.codes
        line["color"] = values
        line["colorscale"] = [
            colors[index % len(colors)] for index in range(len(np.unique(values)))
        ]
    return line


def plot_parallel_categories(
    df,
    target=None,
    columns=None,
    labels={},
    width=None,
    height=None,
    title=None,
    colors=0,
    font={},
    tick_font={},
    axis_font={},
    title_font={},
):
    font = apply_default(get_font(), font)
    #
    colors = get_colors(colors)
    columns = get_categorical_columns(df, columns)
    if target is not None and target not in columns:
        columns = [target] + columns
    #
    if target is not None:
        if target in columns:
            columns.remove(target)
        columns = [target] + columns
    fig = go.Figure(
        go.Parcats(
            dimensions=list(
                [
                    dict(
                        values=df[column],
                        label=labels.get(column, column),
                        categoryorder="category ascending",
                    )
                    for column in columns
                ]
            ),
            line=get_line(df, target, colors),
            labelfont=apply_default(font, axis_font),
            tickfont=apply_default(font, tick_font),
        )
    )
    fig = update_width_and_height(fig, width, height)
    fig = update_title(fig, title, title_font, font)
    # if template is not None:
    #     fig.update_layout(template = template)
    fig.update_layout(font=font)
    return fig
